@@ -20,7 +20,7 @@ from base64 import b64encode, b64decode
 from fair_research_login.client import NativeClient
 from mdml_client.config import CLIENT_ID
 from uuid import uuid4
-from confluent_kafka import Consumer, Producer
+from confluent_kafka import Consumer, Producer, TopicPartition
 from confluent_kafka import SerializingProducer
 from confluent_kafka.serialization import StringSerializer
 from confluent_kafka.schema_registry import SchemaRegistryClient
@@ -197,6 +197,160 @@ py_type_to_schema_type = {
     list: "array"
 }
 
+def start_experiment(id, topics):
+    """
+    Create an experiment that writes all messages from the given topics to a new singular topic
+    
+    Parameters
+    ==========
+    id : str
+        Unique ID for the experiment
+    topics : list(str)
+        Topics to consume from
+    """
+    experiment_topics_schema = {
+        "$schema": "http://merf.egs.anl.gov/mdml-experiment-service-schema#",
+        "title": "ExperimentServiceSchema",
+        "description": "Schema for Kafka MDML Experiments",
+        "type": "object",
+        "properties": {
+            "time": {
+                "description": "Sent timestamp",
+                "type": "number"
+            },
+            "experiment_id": {
+                "description": "A unique experiment ID",
+                "type": "string"
+            },
+            "topics": {
+                "description": "Topics under the experiment",
+                "type": "array",
+                "items": {
+                    "type": "string"
+                },
+            },
+            "status": {
+                "description": "Experiment status",
+                "type": "string"
+            }
+        },
+        "required": [ "time", "experiment_id", "status" ]
+    }
+    producer = kafka_mdml_producer("mdml-experiment-service", schema=experiment_topics_schema)
+    producer.produce({
+        "time": time.time(),
+        "experiment_id": id,
+        "topics": topics,
+        "status": "on"
+    })
+    producer.flush()
+    time.sleep(5)
+    print("Experiment started")
+
+def stop_experiment(id):
+    """
+    Create an experiment that writes all messages from the given topics to a new singular topic
+    
+    Parameters
+    ==========
+    id : str
+        Unique ID for the experiment
+    topics : list(str)
+        Topics to consume from
+    """
+    experiment_topics_schema = {
+        "$schema": "http://merf.egs.anl.gov/mdml-experiment-service-schema#",
+        "title": "ExperimentServiceSchema",
+        "description": "Schema for Kafka MDML Experiments",
+        "type": "object",
+        "properties": {
+            "time": {
+                "description": "Sent timestamp",
+                "type": "number"
+            },
+            "experiment_id": {
+                "description": "A unique experiment ID",
+                "type": "string"
+            },
+            "topics": {
+                "description": "Topics under the experiment",
+                "type": "array",
+                "items": {
+                    "type": "string"
+                },
+            },
+            "status": {
+                "description": "Experiment status",
+                "type": "string"
+            }
+        },
+        "required": [ "time", "experiment_id", "status" ]
+    }
+    producer = kafka_mdml_producer("mdml-experiment-service", schema=experiment_topics_schema)
+    producer.produce({
+        "time": time.time(),
+        "experiment_id": id,
+        "status": "off"
+    })
+    producer.flush()
+    print("Experiment stopped")
+
+def experiment_spawner():
+    experiments={}
+    exp_consumer = kafka_mdml_consumer(["mdml-experiment-service"], "mdml-experiment-service")
+    for msg in exp_consumer.consume(overall_timeout=-1):
+        print(msg)
+        if msg['value']['status'] == "on":
+            #Message will contain topics to include under an experiment
+            t = threading.Thread(target=experiment_task, args=(msg,))
+            experiments[msg['value']['experiment_id']] = t
+            t.setDaemon(True)
+            t.start()
+
+def experiment_task(msg):
+    exp_id = msg['value']['experiment_id']
+    topics = msg['value']['topics']
+    topics.append("mdml-experiment-service")
+    exp_producer = kafka_mdml_producer_schemaless(f"mdml-experiment-{exp_id}")
+    exp_consumer = kafka_mdml_consumer(topics, f"mdml-experiment-{exp_id}", auto_offset_reset="latest")
+    num_msgs = 0
+    for msg in exp_consumer.consume(overall_timeout=-1, verbose=False):
+        if msg['topic'] == "mdml-experiment-service":
+            if msg['value']['status'] == "off":
+                exp_producer.flush()
+                print("Stopping consumer loop...")
+                break
+        num_msgs += 1
+        exp_producer.produce(json.dumps(msg))
+        # Determine optimal strategy for flushing 
+        if num_msgs % 500 == 0:
+            exp_producer.flush()
+
+def replay_experiment(exp_id, group, print_all_data=False):
+    data = []
+    exp_consumer = kafka_mdml_consumer_schemaless([f"mdml-experiment-{exp_id}"], group)
+    print("Gathering experiment data for replaying.")
+    for msg in exp_consumer.consume(overall_timeout=5, verbose=False):
+        data.append(json.loads(msg['value']))
+    if len(data) == 0:
+        print("No experiment data found. You must use a unique group ID for each replay.")
+        return
+    print("Calculating simulation timeline.")
+    for d in data:
+        d['time'] = d['value']['time']
+    data = sorted(data, key=lambda k: k['time'])
+    real_start_timestamp = data[0]['time']
+    for d in data:
+        d['time_diff'] = d['time'] - real_start_timestamp
+    if print_all_data:
+        print(data)
+    print("Starting replay.")
+    sim_start_timestamp = time.time()
+    while len(data) > 0:
+        if data[0]['time_diff'] <= time.time() - sim_start_timestamp:
+            print(data[0])
+            del data[0]
+
 def create_schema(d, title, descr, required_keys=None):
     """
     Input data object is turned into a schema for use
@@ -206,12 +360,12 @@ def create_schema(d, title, descr, required_keys=None):
     ----------
     d : dict
         Data object to translate into a schema
-    required_keys : list of str
-        List of strings of the keys that are required in the schema
     title : str
         Title of the schema
     descr : str
         Description of the schema
+    required_keys : list of str
+        List of strings of the keys that are required in the schema
     """
     schema = {
         "$schema": f"http://merf.egs.anl.gov/mdml-{title}-auto-schema#",
@@ -241,6 +395,17 @@ def create_schema(d, title, descr, required_keys=None):
             }
     return schema
 
+def log_topics(topics, group, output_filepath, verbose=False):
+    """
+    Create an experiment that records the offsets of given topics
+    """
+    consumer = kafka_mdml_consumer(topics, group=group)
+    with open(output_filepath, 'w') as f:
+        for msg in consumer.consume():
+            f.write(json.dumps(msg))
+            if verbose:
+                print(f"Logged message on topic: {msg['topic']}")
+
 class kafka_mdml_producer_schemaless:
     """
     Creates a Producer instance for interacting with the MDML
@@ -268,7 +433,7 @@ class kafka_mdml_producer_schemaless:
             'bootstrap.servers': f'{kafka_host}:{kafka_port}'
         }
         self.producer = Producer(producer_config)
-    def produce(self, data):
+    def produce(self, data, key=None, partition=None):
         """
         Produce data to the supplied topic 
 
@@ -277,7 +442,10 @@ class kafka_mdml_producer_schemaless:
         data : dict
             Dictionary of the data
         """
-        self.producer.produce(topic=self.topic, value=data)
+        if partition is None:
+            self.producer.produce(topic=self.topic, value=data, key=key)
+        else:
+            self.producer.produce(topic=self.topic, value=data, key=key, partition=partition)
     def flush(self):
         """
         Flush (send) any messages currently waiting in the producer.
@@ -335,7 +503,7 @@ class kafka_mdml_producer:
             'value.serializer': json_serializer
         }
         self.producer = SerializingProducer(producer_config)
-    def produce(self, data, key=None):
+    def produce(self, data, key=None, partition=None):
         """
         Produce data to the supplied topic 
 
@@ -343,8 +511,15 @@ class kafka_mdml_producer:
         ----------
         data : dict
             Dictionary of the data
+        key : str
+            String for the Kafka assignor to use to calculate a partition
+        partition : int
+            Number of the partition to assign the message to
         """
-        self.producer.produce(topic=self.topic, key=key, value=data)
+        if partition is None:
+            self.producer.produce(topic=self.topic, value=data, key=key)
+        else:
+            self.producer.produce(topic=self.topic, value=data, key=key, partition=partition)
     def flush(self):
         """
         Flush (send) any messages currently waiting in the producer.
@@ -371,7 +546,7 @@ class kafka_mdml_consumer:
     schema_port : int
         Port of the kafka schema registry
     """
-    def __init__(self, topics, group, 
+    def __init__(self, topics, group, auto_offset_reset="earliest",
                 kafka_host="merf.egs.anl.gov", kafka_port=9092,
                 schema_host="merf.egs.anl.gov", schema_port=8081):
         self.topics = topics
@@ -406,17 +581,21 @@ class kafka_mdml_consumer:
         consumer_conf = {
             'bootstrap.servers': f"{kafka_host}:{kafka_port}",
             'group.id': group,
-            'auto.offset.reset': 'earliest',
+            'auto.offset.reset': auto_offset_reset,
             'allow.auto.create.topics': 'true' # prevents unknown topic error 
         }
         consumer = Consumer(consumer_conf)
         consumer.subscribe(topics)
         self.consumer = consumer
 
-    def consume(self, poll_timeout=1.0, overall_timeout=300.0):
-        print(f"Consumer loop will exit after {overall_timeout} seconds without receiving a message or with Ctrl+C")
+    def consume(self, poll_timeout=1.0, overall_timeout=300.0, verbose=True):
+        if verbose:
+            if overall_timeout != -1:
+                print(f"Consumer loop will exit after {overall_timeout} seconds without receiving a message or with Ctrl+C")
+            else:
+                print(f"Consumer loop will run indefinitely until a Ctrl+C")
         timeout = 0.0
-        while timeout < overall_timeout:
+        while timeout < overall_timeout or overall_timeout == -1:
             try:
                 msg = self.consumer.poll(poll_timeout)
                 if msg is None:
@@ -558,8 +737,9 @@ class kafka_mdml_consumer_schemaless:
         consumer.subscribe(topics)
         self.consumer = consumer
 
-    def consume(self, poll_timeout=1.0, overall_timeout=300.0):
-        print(f"Consumer loop will exit after {overall_timeout} seconds without receiving a message or with Ctrl+C")
+    def consume(self, poll_timeout=1.0, overall_timeout=300.0, verbose=True):
+        if verbose:
+            print(f"Consumer loop will exit after {overall_timeout} seconds without receiving a message or with Ctrl+C")
         timeout = 0.0
         while timeout < overall_timeout:
             try:
